import argparse
import sys  # We need sys so that we can pass argv to QApplication

import numpy as np
import pandas as pd
import warnings
import petab.C as ptc
from PySide2 import QtWidgets, QtCore, QtGui
from PySide2.QtWidgets import QVBoxLayout, QComboBox, QWidget, QLabel
from petab import measurements, core
import pyqtgraph as pg

from . import utils
from . import visuSpec_plot
from . import bar_plot
from . import window_functionality


class MainWindow(QtWidgets.QMainWindow):
    """
    The main window

    Attributes:
        exp_data: PEtab measurement table
        visualization_df: PEtab visualization table
        yaml_dict: Dictionary of the files in the yaml file
        condition_df: PEtab condition table
        observable_df: PEtab observable table
        plot1_widget: pg.GraphicsLayoutWidget containing the main plot
        plot2_widget: pg.GraphicsLayoutWidget containing the correlation plot
        warn_msg: QLabel displaying current warning messages
        table_window: Popup TableWidget displaying the clicked table
        tree_view: QTreeView of the yaml file
        visu_spec_plots: A list of VisuSpecPlots
        cbox: A dropdown menu for the plots
        current_list_index: List index of the currently displayed plot
        wid: QSplitter between main plot and correlation plot
    """
    def __init__(self, exp_data: pd.DataFrame,
                 visualization_df: pd.DataFrame = None,
                 simulation_df: pd.DataFrame = None,
                 condition_df: pd.DataFrame = None,
                 observable_df: pd.DataFrame = None, *args, **kwargs):

        super(MainWindow, self).__init__(*args, **kwargs)
        # set the background color to white
        pg.setConfigOption('background', 'w')
        pg.setConfigOption('foreground', 'k')
        pg.setConfigOption("antialias", True)
        self.resize(1000, 600)
        self.setWindowTitle("PEtab-vis")
        self.yaml_dict = None
        self.visualization_df = visualization_df
        self.simulation_df = simulation_df
        self.condition_df = condition_df
        self.observable_df = observable_df
        self.exp_data = exp_data
        self.visu_spec_plots = []
        self.wid = QtWidgets.QSplitter()
        self.plot1_widget = pg.GraphicsLayoutWidget(show=True)
        self.plot2_widget = pg.GraphicsLayoutWidget(show=False)
        self.wid.addWidget(self.plot1_widget)
        # plot2_widget will be added to the QSplitter when
        # a simulation file is opened
        self.cbox = QComboBox()  # dropdown menu to select plots
        self.cbox.currentIndexChanged.connect(lambda x: self.index_changed(x))
        self.warn_msg = QLabel("")
        # The new window that pops up to display a table
        self.table_window = None
        self.tree_view = QtGui.QTreeView(self)
        self.tree_view.setHeaderHidden(True)
        self.wid.addWidget(self.tree_view)
        self.current_list_index = 0

        warnings.showwarning = self.redirect_warning

        window_functionality.add_file_selector(self)
        if self.exp_data is not None:
            self.add_plots()

        # the layout of the plot-list and message textbox
        lower_layout = QVBoxLayout()
        lower_layout.addWidget(self.cbox)
        lower_layout.addWidget(self.warn_msg)
        lower_widget = QWidget()
        lower_widget.setLayout(lower_layout)
        split_plots_and_warnings = QtWidgets.QSplitter()
        split_plots_and_warnings.setOrientation(QtCore.Qt.Vertical)
        split_plots_and_warnings.addWidget(self.wid)
        split_plots_and_warnings.addWidget(lower_widget)

        layout = QVBoxLayout()
        layout.addWidget(split_plots_and_warnings)

        widget = QWidget()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def add_plots(self):
        """
        Adds the current visuSpecPlots to the main window,
        removes the old ones and updates the
        cbox (dropdown list)

        Returns:
            List of PlotItem
        """
        self.clear_QSplitter()
        self.visu_spec_plots.clear()

        if self.visualization_df is not None:
            # to keep the order of plots consistent with names from the plot selection
            indexes = np.unique(self.visualization_df[ptc.PLOT_ID], return_index=True)[1]
            plot_ids = [self.visualization_df[ptc.PLOT_ID][index] for index in sorted(indexes)]
            for plot_id in plot_ids:
                self.create_and_add_visuPlot(plot_id)

        else:  # default plot when no visu_df is provided
            self.create_and_add_visuPlot()

        plots = [visuPlot.getPlot() for visuPlot in self.visu_spec_plots]

        # update the cbox
        self.cbox.clear()
        # calling this method sets the index of the cbox to 0
        # and thus displays the first plot
        utils.add_plotnames_to_cbox(self.exp_data, self.visualization_df, self.cbox)

        return plots

    def index_changed(self, i: int):
        """
        Changes the displayed plot to the one selected in the dropdown list

        Arguments:
            i: index of the selected plot
        """
        if 0 <= i < len(self.visu_spec_plots):  # i is -1 when the cbox is cleared
            self.clear_QSplitter()
            self.plot1_widget.addItem(self.visu_spec_plots[i].getPlot())
            self.plot2_widget.hide()
            if self.simulation_df is not None:
                self.plot2_widget.show()
                self.plot2_widget.addItem(self.visu_spec_plots[i].correlation_plot)
            self.current_list_index = i

    def keyPressEvent(self, ev):
        """
        Changes the displayed plot by pressing arrow keys

        Arguments:
            ev: key event
        """
        # Exit when pressing ctrl + Q
        ctrl = False
        if (ev.modifiers() & QtCore.Qt.ControlModifier):
            ctrl = True
        if ctrl and ev.key() == QtCore.Qt.Key_Q:
            sys.exit()

        if(ev.key() == QtCore.Qt.Key_Up):
            self.index_changed(self.current_list_index - 1)
        if(ev.key() == QtCore.Qt.Key_Down):
            self.index_changed(self.current_list_index + 1)
        if(ev.key() == QtCore.Qt.Key_Left):
            self.index_changed(self.current_list_index - 1)
        if(ev.key() == QtCore.Qt.Key_Right):
            self.index_changed(self.current_list_index + 1)

    def add_warning(self, message: str):
        """
        Adds the message to the warnings box

        Arguments:
            message: The message to display
        """
        if message not in self.warn_msg.text():
            self.warn_msg.setText(self.warn_msg.text() + message + "\n")

    def redirect_warning(self, message, category, filename=None, lineno=None, file=None, line=None):
        """
        Redirect all warning messages and display them in the window.

        Arguments:
            message: The message of the warning
        """
        print("Warning redirected: " + str(message))
        self.add_warning(str(message))

    def create_and_add_visuPlot(self, plot_id=""):
        """
        Create a visuSpec_plot object based on the given plot_id.
        If no plot_it is provided the default will be plotted.
        Add all the warnings of the visuPlot object to the warning text box.

        The actual plotting happens in the index_changed method

        Arguments:
            plot_id: The plotId of the plot
        """
        # split the measurement df by observable when using default plots
        if self.visualization_df is None:
            # to keep the order of plots consistent with names from the plot selection
            indexes = np.unique(self.exp_data[ptc.OBSERVABLE_ID], return_index=True)[1]
            observable_ids = [self.exp_data[ptc.OBSERVABLE_ID][index] for index in sorted(indexes)]
            for observable_id in observable_ids:
                rows = self.exp_data[ptc.OBSERVABLE_ID] == observable_id
                data = self.exp_data[rows]
                visuPlot = visuSpec_plot.VisuSpecPlot(measurement_df=data, visualization_df=None,
                                                      condition_df=self.condition_df,
                                                      simulation_df=self.simulation_df, plotId=plot_id)
                self.visu_spec_plots.append(visuPlot)
                if visuPlot.warnings:
                    self.add_warning(visuPlot.warnings)

        else:
            # reduce the visualization df to the relevant rows (by plotId)
            rows = self.visualization_df[ptc.PLOT_ID] == plot_id
            visu_df = self.visualization_df[rows]
            if ptc.PLOT_TYPE_SIMULATION in visu_df.columns and\
                    visu_df.iloc[0][ptc.PLOT_TYPE_SIMULATION] == ptc.BAR_PLOT:
                barPlot = bar_plot.BarPlot(measurement_df=self.exp_data,
                                           visualization_df=visu_df,
                                           condition_df=self.condition_df,
                                           simulation_df=self.simulation_df, plotId=plot_id)
                # might want to change the name of visu_spec_plots to clarify that
                # it can also include bar plots (maybe to plots?)
                self.visu_spec_plots.append(barPlot)
            else:
                visuPlot = visuSpec_plot.VisuSpecPlot(measurement_df=self.exp_data,
                                                      visualization_df=visu_df,
                                                      condition_df=self.condition_df,
                                                      simulation_df=self.simulation_df, plotId=plot_id)
                self.visu_spec_plots.append(visuPlot)
                if visuPlot.warnings:
                    self.add_warning(visuPlot.warnings)

    def clear_QSplitter(self):
        """
        Clear the GraphicsLayoutWidgets for the
        measurement and correlation plot
        """
        self.plot1_widget.clear()
        self.plot2_widget.clear()


def main():
    options = argparse.ArgumentParser()
    options.add_argument("-m", "--measurement", type=str, required=False,
                         help="PEtab measurement file", default=None)
    options.add_argument("-v", "--visualization", type=str, required=False,
                         help="PEtab visualization file", default=None)
    args = options.parse_args()

    exp_data = None
    if args.measurement is not None:
        exp_data = measurements.get_measurement_df(args.measurement)

    visualization_df = None
    if args.visualization is not None:
        visualization_df = core.concat_tables(args.visualization, core.get_visualization_df)

    app = QtWidgets.QApplication(sys.argv)
    main = MainWindow(exp_data, visualization_df)
    main.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
import petab.C as ptc
from PySide2 import QtCore
import pyqtgraph as pg

from . import plot_row
from . import utils
from . import plot_class


class VisuSpecPlot(plot_class.PlotClass):
    """
    Can generate a line plot based on the given specifications

    Arguments:
        measurement_df: PEtab measurement table
        visualization_df: PEtab visualization table
        simulation_df: PEtab simulation table
        condition_df: PEtab condition table
        plotId: Id of the plot (has to in the visualization_df aswell)

    Attributes:
        scatter_points: A dictionary containing 2 lists for
            the x- and y-values respectively
        scatter_points_simulation: A dictionary containing 2 lists for
            the x- and y-values respectively
        plot_rows: A list of PlotRows
        plot_rows_simulation: A list of PlotRows for simulation data
        exp_lines: A list of PlotDataItems
        simu_lines: A list of PlotDataItems for simulation data
    """
    def __init__(self, measurement_df: pd.DataFrame = None,
                 visualization_df: pd.DataFrame = None,
                 simulation_df: pd.DataFrame = None,
                 condition_df: pd.DataFrame = None,
                 plotId: str = ""):
        super().__init__(measurement_df, visualization_df, simulation_df,
                         condition_df, plotId)

        # reduce the visualization_df to the relevant rows (by plotId)
        if self.visualization_df is not None:
            # Note the visualization df is already reduced
            # before creating the visuSpecPlot object
            self.check_log_for_zeros()

        self.legend = self.plot.addLegend()

        # useful to remove them from the plot when disabling lines
        self.datasetId_to_plotDataItem = {}
        self.datasetId_to_errorbar = {}
        self.datasetId_to_points = {}

        self.plot_rows = []  # list of plot_rows
        self.plot_rows_simulation = []
        self.overview_df = pd.DataFrame(columns=["x", "y", "name", "is_simulation", "dataset_id", "x_var"])
        self.exp_lines = []  # list of PlotDataItems (measurements)
        self.simu_lines = []  # (simulations)

        self.plot_everything()

    def plot_everything(self):
        """
        Generate the list of plotRows (one for each line in the visualization file).
        Create the overview_df based on the plotRows with respect to disabled rows.
        Create a list of plotDataItems for each plotRow.
        Generate the plot based on the plotDataItems.
        Generate the correlation plot if a simulation file is provided based on the overview df
        """
        self.plot.clear()
        self.error_bars = []
        self.plot_rows = self.generate_plot_rows(self.measurement_df)  # list of plot_rows
        self.plot_rows_simulation = self.generate_plot_rows(self.simulation_df)
        self.overview_df = self.generate_overview_df()

        self.exp_lines = self.generate_plot_data_items(self.plot_rows,
                                                       is_simulation=False)  # list of PlotDataItems (measurements)
        self.simu_lines = self.generate_plot_data_items(self.plot_rows_simulation, is_simulation=True)  # (simulations)

        # make sure the is_simulation column is really boolean because otherwise
        # the logical not operator ~ causes problems
        self.overview_df["is_simulation"] = self.overview_df["is_simulation"].astype("bool")
        self.generate_plot()

        if self.simulation_df is not None:
            # add the correlation plot (only if a simulation file is provided)
            # inherited method from PlotClass
            self.generate_correlation_plot(self.overview_df)

    def generate_overview_df(self):
        """
        Generate the overview df containing the x- and y-data, the name,
        the dataset_id, the x_var and simulation information of all enabled
        plotRows.

        Returns:
            overview_df: A dataframe containing an overview of the plotRows
        """
        overview_df = pd.DataFrame(columns=["x", "y", "name", "is_simulation", "dataset_id", "x_label"])
        if self.visualization_df is not None:
            dfs = [p_row.get_data_df() for p_row in (self.plot_rows + self.plot_rows_simulation)
                   if p_row.dataset_id not in self.disabled_rows]
            if dfs:
                overview_df = pd.concat(dfs, ignore_index=True)
        return overview_df

    def generate_plot_rows(self, df):
        """
        Create a PlotRow object for each row of the visualization df

        Arguments:
            df: Measurement or Simulation df
        """
        plot_rows = []
        if self.visualization_df is not None:
            for _, plot_spec in self.visualization_df.iterrows():
                if df is not None:
                    plot_line = plot_row.PlotRow(df, plot_spec, self.condition_df)
                    plot_rows.append(plot_line)
        return plot_rows

    def generate_plot_data_items(self, plot_rows, is_simulation: bool = False):
        """
        Generate a list of PlotDataItems based on
        a list of PlotRows

        Arguments:
            plot_rows: A list of PlotRow objects
            is_simulation: True plot_rows belong to a simulation df
        Returns:
            pdis: A list of PlotDataItems
        """
        pdis = []  # list of PlotDataItems
        for line in plot_rows:
            if line.dataset_id == "":
                plot_lines = self.default_plot(line, is_simulation=is_simulation)
                pdis = pdis + plot_lines
            else:
                if line.dataset_id not in self.disabled_rows:
                    pdis.append(self.plot_row_to_plot_data_item(line))
        return pdis

    def generate_plot(self):
        """
        Generate a pyqtgraph PlotItem based on exp_lines
        and simu_lines (both are DataPlotItem lists

        Returns:
            pyqtgraph PlotItem
        """
        if len(self.plot_rows) > 0:
            # get the axis labels info from the first line of the plot
            self.plot.setLabel("left", self.plot_rows[0].y_label)
            self.plot.setLabel("bottom", self.plot_rows[0].x_label)

        else:  # when no visualization file was provided
            self.plot.setLabel("left", "measurement")
            self.plot.setLabel("bottom", "time")
            self.exp_lines = self.default_plot(None)
            if self.simulation_df is not None:
                self.simu_lines = self.default_plot(None, is_simulation=True)

        # color the plot so measurements and simulations
        # have the same color but are different from other
        # measurements
        num_lines = len(self.plot_rows)
        for i, line in enumerate(self.exp_lines):
            color = pg.intColor(i, hues=num_lines)
            line.setPen(color, style=QtCore.Qt.DashDotLine, width=2)
            self.plot.addItem(line)
            if len(self.simu_lines) > 0:
                self.simu_lines[i].setPen(color, style=QtCore.Qt.SolidLine, width=2)
                self.plot.addItem(self.simu_lines[i])

        self.add_measurements_points()

        # Errorbars do not support log scales
        if self.plot_rows and ("log" in self.plot_rows[0].x_scale or "log" in self.plot_rows[0].y_scale):
            if len(self.error_bars) > 0:
                self.warnings = self.warnings + "Errorbars are not supported with log scales (in " \
                                + self.plot_title + ")\n"
        else:
            # add error bars
            for error_bar in self.error_bars:
                self.plot.addItem(error_bar)

        self.set_scales()

        return self.plot

    def add_measurements_points(self):
        """
        Add the measurement points to the plot
        """
        dataset_ids = np.unique(self.overview_df["dataset_id"])
        for id in dataset_ids:
            df = self.overview_df[self.overview_df["dataset_id"] == id]
            x = df[~df["is_simulation"]]["x"].tolist()
            measurements = df[~df["is_simulation"]]["y"].tolist()
            points = self.plot.plot(x, measurements,
                                    pen=None, symbol='o',
                                    symbolBrush=pg.mkBrush(0, 0, 0), symbolSize=6)
            self.datasetId_to_points[id] = points
            x_simulation = df[df["is_simulation"]]["x"].tolist()
            simulations = df[df["is_simulation"]]["y"].tolist()
            points = self.plot.plot(x_simulation, simulations,
                                    pen=None, symbol='o',
                                    symbolBrush=pg.mkBrush(255, 255, 255), symbolSize=6)
            self.datasetId_to_points[id + "_simulation"] = points

    def plot_row_to_plot_data_item(self, p_row: plot_row.PlotRow):
        """
        Creates a PlotDataItem based on the PlotRow.
        Also, generate an error bar and measurement points.

        Arguments:
            p_row: The PlotRow object that contains the information
             of the line that is added
        """
        legend_name = p_row.legend_name
        if p_row.is_simulation:
            legend_name = legend_name + " simulation"
        pdi = pg.PlotDataItem(p_row.x_data, p_row.y_data, name=legend_name)
        # add it to the dict (used for disabling rows by dataset_id)
        if p_row.is_simulation:
            self.datasetId_to_plotDataItem[p_row.dataset_id + "_simulation"] = pdi
        else:
            self.datasetId_to_plotDataItem[p_row.dataset_id] = pdi

        # Only add error bars when needed
        if (p_row.has_replicates or p_row.plot_type_data == ptc.PROVIDED)\
                and p_row.plot_type_data != ptc.REPLICATE:
            error_length = p_row.sd
            if p_row.plot_type_data == ptc.MEAN_AND_SEM:
                error_length = p_row.sem
            if p_row.plot_type_data == ptc.PROVIDED:
                error_length = p_row.provided_noise
            beam_width = 0
            if len(p_row.x_data) > 0:  # p_row.x_data could be empty
                beam_width = np.max(p_row.x_data) / 100
            error = pg.ErrorBarItem(x=p_row.x_data, y=p_row.y_data, top=error_length, bottom=error_length, beam=beam_width)
            self.error_bars.append(error)
            # add it to the dict (used for disabling rows by dataset_id)
            if p_row.is_simulation:
                self.datasetId_to_errorbar[p_row.dataset_id + "_simulation"] = error
            else:
                self.datasetId_to_errorbar[p_row.dataset_id] = error

        return(pdi)

    def default_plot(self, p_row: plot_row.PlotRow, is_simulation=False):
        """
        This method is used when the p_row contains no dataset_id
        or no visualization file was provided
        Therefore, the whole dataset will be visualized
        in a single plot
        The plotDataItems created here will be added to self.exp_lines

        Arguments:
            p_row: The PlotRow object that contains the information
             of the line that is added

        Returns:
            List of Plot_Data_Items
        """
        plot_lines = []
        # group by Observable ID as default
        grouping = ptc.SIMULATION_CONDITION_ID
        # if the datasetId column is present, group by datasetId
        if ptc.DATASET_ID in self.measurement_df.columns:
            grouping = ptc.DATASET_ID
        else:
            self.add_warning("Grouped by observable. If you want to specify another grouping option"
                             ", please add \"datasetID\" columns.")
        df = self.measurement_df
        y_var = ptc.MEASUREMENT
        if is_simulation:
            df = self.simulation_df
            y_var = ptc.SIMULATION

        for group_id in np.unique(df[grouping]):
            line_data = df[df[grouping] == group_id]
            data = line_data[[y_var, ptc.TIME]]
            x_data = data.groupby(ptc.TIME)
            x_data = np.fromiter(x_data.groups.keys(), dtype=float)
            y_data = utils.mean_replicates(line_data, ptc.TIME, y_var)
            line_name = group_id

            # case distinction if a visualization_df was provided or not
            if p_row is not None:
                # add offsets to the data:
                x_data = x_data + p_row.x_offset
                y_data = y_data + p_row.y_offset
            else:
                line_name = line_name + "_" + df[ptc.OBSERVABLE_ID].iloc[0]
            # add points
            if is_simulation:
                line_name = line_name + " simulation"
                line_df = pd.DataFrame({"x": x_data.tolist(), "y": y_data.tolist(), "name": group_id, "is_simulation": True})
            else:
                line_df = pd.DataFrame({"x": x_data.tolist(), "y": y_data.tolist(), "name": group_id, "is_simulation": False})
            self.overview_df = self.overview_df.append(line_df, ignore_index=True)

            plot_lines.append(pg.PlotDataItem(x_data, y_data, name=line_name))

        return plot_lines

    def set_scales(self):
        """
        Set the scales to log10 if necessary.
        Default is linear scale.
        """
        if len(self.plot_rows) > 0:  # default plots have a linear scale
            if "log" in self.plot_rows[0].x_scale:
                self.plot.setLogMode(x=True)
                if self.plot_rows[0].x_scale == "log":
                    self.add_warning("log not supported, using log10 instead (in " + self.plot_title + ")")
            if "log" in self.plot_rows[0].y_scale:
                self.plot.setLogMode(y=True)
                if self.plot_rows[0].y_scale == "log":
                    self.add_warning("log not supported, using log10 instead (in " + self.plot_title + ")")

    def check_log_for_zeros(self):
        """
        Add an offset to values if they contain a zero and will be plotted on
        log-scale.
        The offset is calculated as the smalles nonzero value times 0.001
        (Also adds the offset to the simulation values).
        """
        x_var = utils.get_x_var(self.visualization_df.iloc[0])
        y_var = ptc.MEASUREMENT
        if x_var == ptc.TIME:
            x_values = np.asarray(self.measurement_df[x_var])
        else:
            # for concentration plots, each line can have a
            # different x_var
            x_values = []
            for variable in self.visualization_df[ptc.X_VALUES]:
                x_values = x_values + list(self.condition_df[variable])
            x_values = np.asarray(x_values)

        y_values = np.asarray(self.measurement_df[y_var])

        if ptc.X_SCALE in self.visualization_df.columns:
            if 0 in x_values and "log" in self.visualization_df.iloc[0][ptc.X_SCALE]:
                offset = np.min(x_values[np.nonzero(x_values)]) * 0.001
                if x_var == ptc.TIME:
                    x_values = x_values + offset
                    self.measurement_df[x_var] = x_values
                else:
                    for variable in self.visualization_df[ptc.X_VALUES]:
                        self.condition_df[variable] = np.asarray(self.condition_df[variable]) + offset
                self.add_warning("Unable to take log of 0, added offset of " + str(offset) + " to x-values")

                if self.simulation_df is not None:
                    x_simulation = np.asarray(self.simulation_df[x_var])
                    self.simulation_df[x_var] = x_simulation + offset

        if ptc.Y_SCALE in self.visualization_df.columns:
            if 0 in y_values and "log" in self.visualization_df.iloc[0][ptc.Y_SCALE]:
                offset = np.min(y_values[np.nonzero(y_values)]) * 0.001
                y_values = y_values + offset
                self.measurement_df[y_var] = y_values
                self.add_warning("Unable to take log of 0, added offset of " + str(offset) + " to y-values")

                if self.simulation_df is not None:
                    y_simulation = np.asarray(self.simulation_df[ptc.SIMULATION])
                    self.simulation_df[ptc.SIMULATION] = y_simulation + offset

    def add_or_remove_line(self, dataset_id):
        """
        Add the datasetId to the disabled rows if
        it is not in the disabled rows set. Otherwise,
        remove it from the disabled rows set.
        Then, adjust the plot showing only the enabled rows.

        Arguments:
            dataset_id: The datasetId of the row that should
                        be added or removed.
        """
        # TODO: generate warning for rows without dataset id
        if dataset_id in self.disabled_rows:
            self.disabled_rows.remove(dataset_id)
            self.enable_line(dataset_id)
            if self.simulation_df is not None:
                self.enable_line(dataset_id + "_simulation")
        else:
            self.disabled_rows.add(dataset_id)
            self.disable_line(dataset_id)
            if self.simulation_df is not None:
                self.disable_line(dataset_id + "_simulation")

        # update the correlation plot
        if self.simulation_df is not None:
            overview_df = self.generate_overview_df()
            self.generate_correlation_plot(overview_df)

    def disable_line(self, dataset_id):
        """
        Remove the line from the plot matching the dataset id.
        Also removes the line of the simulation if present.

        Arguments:
            dataset_id: The dataset id of the line that should be removed.
        """
        self.plot.removeItem(self.datasetId_to_plotDataItem[dataset_id])
        if self.datasetId_to_errorbar:  # The plot may not have errorbars
            self.plot.removeItem(self.datasetId_to_errorbar[dataset_id])
        self.plot.removeItem(self.datasetId_to_points[dataset_id])

    def enable_line(self, dataset_id):
        """
        Add the line to the plot matching the dataset id.
        Also add the simulation line if possible.

        Arguments:
            dataset_id: The dataset id of the line that should be added.
        """
        self.plot.addItem(self.datasetId_to_plotDataItem[dataset_id])
        if self.datasetId_to_errorbar:  # The plot may not have errorbars
            self.plot.addItem(self.datasetId_to_errorbar[dataset_id])
        self.plot.addItem(self.datasetId_to_points[dataset_id])



# Import after PySide2 to ensure usage of correct Qt library
import os
import sys
from pathlib import Path

import pandas as pd
from PySide2 import QtWidgets, QtCore, QtGui
from PySide2.QtCore import (QAbstractTableModel, QModelIndex, Qt,
                            QSortFilterProxyModel)
from PySide2.QtGui import QColor
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (QAction, QVBoxLayout, QHeaderView,
                               QSizePolicy, QTableView, QWidget, QFileDialog)

import petab
import petab.C as ptc
from petab import core
from petab.visualize.helper_functions import check_ex_exp_columns


class CustomTableModel(QAbstractTableModel):
    """PEtab data table model."""

    def __init__(self, df=None):
        QAbstractTableModel.__init__(self)
        self.load_data(df)
        self.df = df

    def load_data(self, data):
        for x in data:
            setattr(self, x, data[x])
        self.column_count = data.shape[1]
        self.row_count = data.shape[0]

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        row = index.row()
        column = index.column()
        self.df.iloc[row, column] = value
        return True

    def rowCount(self, parent=QModelIndex()):
        return self.row_count

    def columnCount(self, parent=QModelIndex()):
        return self.column_count

    def headerData(self, section, orientation, role=None):
        if role != Qt.DisplayRole:
            return None

        if orientation == Qt.Horizontal:
            return self.df.columns[section]
        else:
            return self.df.index[section]

    def data(self, index, role=Qt.DisplayRole):
        column = index.column()
        row = index.row()

        if role == Qt.DisplayRole:
            return str(self.df.iloc[row, column])

        elif role == Qt.BackgroundRole:
            return QColor(Qt.white)

        elif role == Qt.TextAlignmentRole:
            return Qt.AlignRight

        return None

    def get_value(self, row, column):
        return self.df.iloc[row][column]


class VisualizaionTableModel(CustomTableModel):
    """
    Special table model for visualization files.
    Make the first column of the table editable for
    the checkbox column.
    Highlight the rows of the currently displayed plot.
    """
    def __init__(self, df=None, window=None):
        CustomTableModel.__init__(self, df)
        self.window = window

    def flags(self, index):
        if not index.isValid():
            return 0

        if index.column() == 0:
            return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsUserCheckable

        return Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.BackgroundRole:
            current_plot = self.window.visu_spec_plots[self.window.current_list_index]
            current_plot_id = current_plot.plotId
            if self.df[ptc.PLOT_ID][index.row()] == current_plot_id:
                return QtGui.QColor("yellow")
        else:
            return super().data(index, role)

    def get_window(self):
        return self.window


class CheckBoxDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QCheckBox cell to the column to which it's applied.
    Used for the visualization table to add the checkbox column and provide it's functionality.
    """
    def __init__(self, parent):
        QtWidgets.QItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):
        """
        Important, otherwise an editor is created if the user clicks in this cell.
        """
        return None

    def paint(self, painter, option, index):
        """
        Paint a checkbox without the label.
        """
        self.drawCheck(painter, option, option.rect, QtCore.Qt.Unchecked if int(index.data()) == 0 else QtCore.Qt.Checked)

    def editorEvent(self, event, model, option, index):
        '''
        Change the data in the model and the state of the checkbox
        if the user presses the left mousebutton and this cell is editable. Otherwise do nothing.
        '''
        if not int(index.flags() & QtCore.Qt.ItemIsEditable) > 0:
            return False

        if event.type() == QtCore.QEvent.MouseButtonRelease and event.button() == QtCore.Qt.LeftButton:
            # Change the checkbox-state
            plotId = model.sourceModel().get_value(index.row(), ptc.PLOT_ID)
            datasetId = model.sourceModel().get_value(index.row(), ptc.DATASET_ID)
            window = model.sourceModel().get_window()
            visu_spec_plot = [visu_spec_plot for visu_spec_plot in window.visu_spec_plots if visu_spec_plot.plotId == plotId][0]
            visu_spec_plot.add_or_remove_line(datasetId)
            self.setModelData(None, model, index)
            return True

        return False

    def setModelData(self, editor, model, index):
        """
        Change the state of the checkbox after it was clicked.
        """
        model.setData(index, 1 if int(index.data()) == 0 else 0, QtCore.Qt.EditRole)


class TableWidget(QWidget):
    """Widget for displaying a PEtab table."""

    def __init__(self, data: pd.DataFrame, add_checkbox_col: bool, window):
        QWidget.__init__(self)

        # Set the Model
        if add_checkbox_col:
            self.model = VisualizaionTableModel(data, window)
        else:  # for any other df
            self.model = CustomTableModel(data)

        # Creating a QTableView
        self.table_view = QTableView()
        self.filter_proxy = QSortFilterProxyModel()
        self.filter_proxy.setSourceModel(self.model)
        self.table_view.setModel(self.filter_proxy)
        self.table_view.setSortingEnabled(True)

        # add a checkbox column for visualization dfs
        if add_checkbox_col:
            delegate = CheckBoxDelegate(None)
            self.table_view.setItemDelegateForColumn(0, delegate)

        # QTableView Headers
        self.horizontal_header = self.table_view.horizontalHeader()
        self.horizontal_header.setSortIndicator(-1, Qt.DescendingOrder)
        self.vertical_header = self.table_view.verticalHeader()
        self.horizontal_header.setSectionResizeMode(QHeaderView.ResizeToContents)

        # QWidget Layout
        self.main_layout = QVBoxLayout()
        size = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)

        size.setHorizontalStretch(1)
        self.table_view.setSizePolicy(size)
        self.main_layout.addWidget(self.table_view)

        self.setLayout(self.main_layout)


def pop_up_table_view(window: QtWidgets.QMainWindow, df: pd.DataFrame):
    """
    Create a popup window that displays the dataframe.

    Arguments:
        window: The main window to which the TableWidget gets added.
        df: The dataframe to display
    """
    add_checkbox_col = False
    if window.visualization_df.equals(df):
        add_checkbox_col = True
    window.table_window = TableWidget(data=df, add_checkbox_col=add_checkbox_col, window=window)
    window.table_window.setGeometry(QtCore.QRect(100, 100, 800, 400))
    window.table_window.show()


def table_tree_view(window: QtWidgets.QMainWindow, folder_path):
    """
    Create a treeview of the yaml file.
    Set the windows df attributes equal to the first file
    in each branch.

    Arguments:
        window: The main window to which the treeview is added
        folder_path: The path to the folder the yaml file is in
    """
    model = QtGui.QStandardItemModel()
    tree_view = window.tree_view
    root_node = model.invisibleRootItem()

    # iterate through the yaml_dict
    for key in window.yaml_dict:
        branch = QtGui.QStandardItem(key)
        is_first_df = True

        # iterate through the files of a yaml_dict entry
        for filename in window.yaml_dict[key]:
            file = QtGui.QStandardItem(filename)
            df = None
            if key == ptc.MEASUREMENT_FILES:
                df = petab.get_measurement_df(folder_path + "/" + filename)
                if is_first_df:
                    window.exp_data = df
            if key == ptc.VISUALIZATION_FILES:
                df = petab.get_visualization_df(folder_path + "/" + filename)
                df.insert(0, "Displayed", 1)  # needed for the checkbox column
                if is_first_df:
                    window.visualization_df = df
            if key == ptc.CONDITION_FILES:
                df = petab.get_condition_df(folder_path + "/" + filename)
                if is_first_df:
                    window.condition_df = df
            if key == ptc.OBSERVABLE_FILES:
                df = petab.get_observable_df(folder_path + "/" + filename)
                if is_first_df:
                    window.observable_df = df
            file.setData(df, role=Qt.UserRole + 1)
            branch.appendRow(file)
            is_first_df = False
        root_node.appendRow(branch)

    if window.simulation_df is not None:
        branch = QtGui.QStandardItem("simulation_files")
        simulation_file = QtGui.QStandardItem("simulation_file")
        simulation_file.setData(window.simulation_df, role=Qt.UserRole + 1)
        branch.appendRow(simulation_file)
        root_node.appendRow(branch)

    tree_view.setModel(model)
    reconnect(tree_view.clicked, lambda i: exchange_dataframe_on_click(i, model, window))
    reconnect(tree_view.doubleClicked, lambda i: display_table_on_doubleclick(i, model, window))


def reconnect(signal, new_function=None):
    """
    Disconnect a signal from all functions and connect it
    to a new function.

    Arguments:
        signal: The signal to reconnect
        new_function: The function the signal gets connected to
    """
    try:
        signal.disconnect()
    except RuntimeError:
        pass
    if new_function is not None:
        signal.connect(new_function)


def exchange_dataframe_on_click(index: QtCore.QModelIndex,
                                model: QtGui.QStandardItemModel,
                                window: QtWidgets.QMainWindow):
    """
    Changes the currently plotted dataframe with the one
    that gets clicked on and replot the data,
    e.g. switch the measurement or visualization df.

    Arguments:
        index: index of the clicked dataframe
        model: model containing the data
        window: Mainwindow whose attributes get updated
    """
    df = model.data(index, role=Qt.UserRole + 1)
    parent = index.parent()
    parent_name = model.data(parent, QtCore.Qt.DisplayRole)
    # Only replot when a new dataframe is selected
    # (Important because double clicking also calls this function)
    df_changed = True
    if df is None:
        return
    if parent_name == ptc.MEASUREMENT_FILES:
        if window.exp_data.equals(df):
            df_changed = False
        window.exp_data = df
    if parent_name == ptc.VISUALIZATION_FILES:
        if window.visualization_df.equals(df):
            df_changed = False
        window.visualization_df = df
    if parent_name == ptc.CONDITION_FILES:
        if window.condition_df.equals(df):
            df_changed = False
        window.condition_df = df
    if parent_name == ptc.OBSERVABLE_FILES:
        if window.observable_df.equals(df):
            df_changed = False
        window.observable_df = df
    if parent_name == "simulation_files":
        if window.simulation_df.equals(df):
            df_changed = False
        window.simulation_df = df

    if df_changed:
        window.add_plots()


def display_table_on_doubleclick(index: QtCore.QModelIndex,
                                 model: QtGui.QStandardItemModel,
                                 window: QtWidgets.QMainWindow):
    """
    Display the dataframe in a new window upon double-click.

    Arguments:
        index: index of the clicked dataframe
        model: model containing the data
        window: Mainwindow whose attributes get updated
    """
    df = model.data(index, role=Qt.UserRole + 1)
    if df is not None:
        pop_up_table_view(window, df)


def add_file_selector(window: QtWidgets.QMainWindow):
    """
    Add a file selector button to the main window.
    Arguments:
        window: Mainwindow
    """
    open_yaml_file = QAction(QIcon('open.png'), 'Open YAML file...', window)
    open_yaml_file.triggered.connect(lambda x: show_yaml_dialog(x, window))
    open_simulation_file = QAction(QIcon('open.png'), 'Open simulation file...', window)
    open_simulation_file.triggered.connect(lambda x: show_simulation_dialog(x, window))
    quit = QAction("Quit", window)
    quit.triggered.connect(sys.exit)

    menubar = window.menuBar()
    fileMenu = menubar.addMenu('&Select File')
    fileMenu.addAction(open_yaml_file)
    fileMenu.addAction(open_simulation_file)
    fileMenu.addAction(quit)


def show_yaml_dialog(self, window: QtWidgets.QMainWindow):
    """
    Display a file selector window when clicking on the select YAML file menu
    item, then display the new plots described by the YAML file.

    Arguments:
        window: Mainwindow
    """
    home_dir = str(Path.home())
    # start file selector on the last selected directory
    settings = QtCore.QSettings("petab", "petabvis")
    if settings.value("last_dir") is not None:
        home_dir = settings.value("last_dir")
    file_name = QFileDialog.getOpenFileName(window, 'Open file', home_dir)[0]
    if file_name != "":  # if a file was selected
        # save the directory for the next use
        last_dir = os.path.dirname(file_name) + "/"
        settings.setValue("last_dir", last_dir)

        window.warn_msg.setText("")

        # select the first df in the dict for measurements, etc.
        yaml_dict = petab.load_yaml(file_name)["problems"][0]
        window.yaml_dict = yaml_dict
        if ptc.VISUALIZATION_FILES not in yaml_dict:
            window.visualization_df = None
            window.add_warning("The YAML file contains no visualization file (default plotted)")
        window.simulation_df = None

        # table_tree_view sets the df attributes of the window
        # equal to the first file of each branch (measurement, visualization, ...)
        window.listWidget = table_tree_view(window, last_dir)
        window.add_plots()


def show_simulation_dialog(self, window: QtWidgets.QMainWindow):
    """
    Display a file selector window when clicking on the select simulation file button,
    then add the simulation lines to the plots.

    Arguments:
        window: main window.
    """
    home_dir = str(Path.home())
    settings = QtCore.QSettings("petab", "petabvis")
    if settings.value("last_dir") is not None:
        home_dir = settings.value("last_dir")
    file_name = QFileDialog.getOpenFileName(window, 'Open simulation file', home_dir)[0]
    if file_name:  # if a file was selected
        if window.exp_data is None:
            window.add_warning("Please open a YAML file first.")
        else:
            window.warn_msg.setText("")
            sim_data = core.get_simulation_df(file_name)
            # check columns, and add non-mandatory default columns
            sim_data, _, _ = check_ex_exp_columns(sim_data, None, None, None, None, None,
                                                  window.condition_df, sim=True)
            # delete the replicateId column if it gets added to the simulation table
            # but is not in exp_data because it causes problems when splitting the replicates
            if ptc.REPLICATE_ID not in window.exp_data.columns and ptc.REPLICATE_ID in sim_data.columns:
                sim_data.drop(ptc.REPLICATE_ID, axis=1, inplace=True)

            if len(window.yaml_dict[ptc.MEASUREMENT_FILES]) > 1:
                window.add_warning("Not Implemented Error: Loading a simulation file with "
                                   "multiple measurement files is currently not supported.")
            else:
                window.simulation_df = sim_data
                window.add_plots()

                # insert correlation plot at position 1
                window.wid.insertWidget(1, window.plot2_widget)
                window.listWidget = table_tree_view(window, os.path.dirname(file_name))

        # save the directory for the next use
        last_dir = os.path.dirname(file_name) + "/"
        settings.setValue("last_dir", last_dir)



import pandas as pd
import pyqtgraph as pg
import petab
import scipy

from . import utils


class PlotClass:
    """
    Arguments:
        measurement_df: PEtab measurement table
        visualization_df: PEtab visualization table
        simulation_df: PEtab simulation table
        condition_df: PEtab condition table
        plotId: Id of the plot (has to in the visualization_df aswell)

    Attributes:
        measurement_df: PEtab measurement table
        visualization_df: PEtab visualization table
        simulation_df: PEtab simulation table
        condition_df: PEtab condition table
        plotId: Id of the plot (has to in the visualization_df aswell)
        error_bars: A list of pg.ErrorBarItems
        warnings: String of warning messages if the input is incorrect
            or not supported
        has_replicates: Boolean, true if replicates are present
        plot_title: The title of the plot
        plot: PlotItem for the main plot (line or bar)
        correlation_plot: PlotItem for the correlation plot
            between measurement and simulation values
    """

    def __init__(self, measurement_df: pd.DataFrame = None,
                 visualization_df: pd.DataFrame = None,
                 simulation_df: pd.DataFrame = None,
                 condition_df: pd.DataFrame = None,
                 plotId: str = ""):

        self.measurement_df = measurement_df
        self.visualization_df = visualization_df
        self.simulation_df = simulation_df
        self.condition_df = condition_df
        self.plotId = plotId
        self.error_bars = []
        self.disabled_rows = set()  # set of plot_ids that are disabled
        self.warnings = ""
        self.has_replicates = petab.measurements.measurements_have_replicates(self.measurement_df)
        self.plot_title = utils.get_plot_title(self.visualization_df)
        self.plot = pg.PlotItem(title=self.plot_title)
        self.correlation_plot = pg.PlotItem(title="Correlation")

    def generate_correlation_plot(self, overview_df):
        """
        Generate the scatterplot between the
        measurement and simulation values.

        Arguments:
            overview_df: Dataframe containing info about the points

        """
        self.correlation_plot.clear()

        if not overview_df.empty:
            measurements = overview_df[~overview_df["is_simulation"]]["y"].tolist()
            simulations = overview_df[overview_df["is_simulation"]]["y"].tolist()

            self.add_points(overview_df)
            self.correlation_plot.setLabel("left", "Simulation")
            self.correlation_plot.setLabel("bottom", "Measurement")

            min_value = min(measurements + simulations)
            max_value = max(measurements + simulations)
            self.correlation_plot.setRange(xRange=(min_value, max_value), yRange=(min_value, max_value))
            self.correlation_plot.addItem(pg.InfiniteLine([0, 0], angle=45))

            # calculate and add the r_squared value
            self.r_squared = self.get_R_squared(measurements, simulations)
            r_squared_text = "R Squared:\n" + str(self.r_squared)[0:5]
            r_squared_text = pg.TextItem(str(r_squared_text), anchor=(0, 0), color="k")
            r_squared_text.setPos(min_value, max_value)
            self.correlation_plot.addItem(r_squared_text, anchor=(0, 0), color="k")

    def add_points(self, overview_df: pd.DataFrame):
        """
        Add the points to the scatterplot and
        display an info text when clicking on a point.

        Arguments:
            overview_df: Dataframe containing info about the points
        """
        # data
        measurements = overview_df[~overview_df["is_simulation"]]["y"].tolist()
        simulations = overview_df[overview_df["is_simulation"]]["y"].tolist()
        names = overview_df[~overview_df["is_simulation"]]["name"].tolist()
        point_descriptions = [(names[i] + "\nmeasurement: " + str(measurements[i]) +
                              "\nsimulation: " + str(simulations[i]))
                              for i in range(len(measurements))]
        # only line plots have x-values, barplots do not
        if "x_label" in overview_df.columns:
            x = overview_df[~overview_df["is_simulation"]]["x"].tolist()
            x_label = overview_df[~overview_df["is_simulation"]]["x_label"].tolist()
            point_descriptions = [(point_descriptions[i] + "\n" + str(x_label[i])) + ": " +
                                  str(x[i]) for i in range(len(point_descriptions))]
        # create the scatterplot
        scatter_plot = pg.ScatterPlotItem(pen=pg.mkPen(None), brush=pg.mkBrush(0, 0, 0))
        spots = [{'pos': [m, s], 'data': idx} for m, s, idx in zip(measurements, simulations, point_descriptions)]
        scatter_plot.addPoints(spots)
        self.correlation_plot.addItem(scatter_plot)

        # add interaction
        last_clicked = None
        info_text = pg.TextItem("", anchor=(0, 0), color="k")
        self.correlation_plot.addItem(info_text)

        def clicked(plot, points):
            nonlocal last_clicked
            nonlocal info_text
            if last_clicked is not None:
                last_clicked.resetPen()
            # remove the text when the same point is clicked twice
            if last_clicked == points[0] and info_text.textItem.toPlainText() != "":
                info_text.setText("")
            else:
                points[0].setPen('b', width=2)
                info_text.setText(str((points[0].data())))
                info_text.setPos(points[0].pos())
                last_clicked = points[0]

        scatter_plot.sigClicked.connect(clicked)

    def get_R_squared(self, measurements, simulations):
        """
        Calculate the R^2 value between the measurement
        and simulation values.

        Arguments:
            measurements: List of measurement values
            simulations: List of simulation values
        Returns:
            The R^2 value
        """
        slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(measurements, simulations)
        print("Linear Regression Statistics for " + self.plot_title + ":")
        print("Slope: " + str(slope) + ", Intercept: " + str(intercept)
              + ", R-value: " + str(r_value) + ", p-value: " + str(p_value)
              + ", Std Err: " + str(std_err))

        return r_value**2

    def add_warning(self, message: str):
        """
        Adds the message to the warnings box

        Arguments:
            message: The message to display
        """
        # filter out double warnings
        if message not in self.warnings:
            self.warnings = self.warnings + message + "\n"

    def getPlot(self):
        return self.plot



import numpy as np
import pandas as pd
import petab.C as ptc

from . import utils
from . import row_class


class PlotRow(row_class.RowClass):
    """
    Can add the content of a visualization_df row to a PlotItem.
    Used for line plots.

    Attributes:
        x_data: Numpy array of the x-values
        y_data: Numpy array of the y-values
        sd: Standard deviation of the replicates
        sem: Standard error of the mean of the replicates
        provided noise: Noise of the measurements
    """
    def __init__(self, exp_data: pd.DataFrame,
                 plot_spec: pd.Series, condition_df: pd.DataFrame, ):

        super().__init__(exp_data, plot_spec, condition_df)

        # calculate new attributes
        self.x_data = self.get_x_data()
        self.y_data = self.get_y_data()
        self.sd = utils.sd_replicates(self.line_data, self.x_var, self.is_simulation)
        self.sem = utils.sem_replicates(self.line_data, self.x_var, self.is_simulation)
        self.provided_noise = self.get_provided_noise()

    def get_x_data(self):
        """
        Return the x-values that should be plotted
        Returns:
            The x-values
        """
        # for concentration plots
        if self.x_var != ptc.TIME:
            x_data = self.condition_df[self.x_var]

        else:  # for time plots
            if self.has_replicates and self.plot_type_data != ptc.REPLICATE:
                # to keep the order intact (only needed if no replicate id col is provided)
                x_data = np.asarray([x_values for x_values, df in self.replicates[0].groupby(self.x_var, sort=True)])
            else:
                x_data = np.asarray(self.replicates[0][self.x_var])
            x_data = x_data + self.x_offset

        return x_data

    def get_y_data(self):
        """
        Return the mean of the y-values that should be plotted if
        the plottype is not ptc.REPLICATE.
        Otherwise, return the y-values of the first replicate.


        Returns:
            The y-values
        """
        variable = self.get_y_variable_name()  # either measurement or simulation
        y_data = np.asarray(self.replicates[0][variable])
        if self.plot_type_data != ptc.REPLICATE:
            y_data = utils.mean_replicates(self.line_data, self.x_var, variable)
        y_data = y_data + self.y_offset

        return y_data

    def get_replicate_x_data(self):
        """
        Return the x-values of each replicate as a list of lists
        Returns:
            x_data_replicates: The y-values for each replicate
        """
        x_data = []
        default_x_values = [df for _, df in self.replicates[0].groupby(self.x_var, sort=True)]
        for replicate in self.replicates:
            if ptc.REPLICATE_ID in self.line_data.columns:
                x_values = np.asarray(replicate[self.x_var])
                # when no explicit replicate id is given, we assume that
                # each replicate uses the same x-values which are determined
                # by the unique x-values in the data
            else:
                x_values = default_x_values
            x_values = x_values + self.x_offset
            x_data.append(x_values)

        return x_data

    def get_replicate_y_data(self):
        """
        Return the y-values of each replicate as a list of lists
        Returns:
            y_data_replicates: The y-values for each replicate
        """
        y_data = []
        variable = self.get_y_variable_name()
        for replicate in self.replicates:
            y_values = np.asarray(replicate[variable])
            y_values = y_values + self.y_offset
            y_data.append(y_values)

        return y_data



import numpy as np
import pandas as pd

from . import row_class


class BarRow(row_class.RowClass):
    """
    Can add the content of a visualization_df row to a PlotItem.
    Used for bar plots.

    Attributes:
        y_data: Y-value
        sd: Standard deviation of the replicates
        sem: Standard error of the mean of the replicates
        provided noise: Noise of the measurements
    """
    def __init__(self, exp_data: pd.DataFrame,
                 plot_spec: pd.Series, condition_df: pd.DataFrame, ):
        super().__init__(exp_data, plot_spec, condition_df)

        # Note: A bar plot has no x_data
        self.y_data = self.get_mean_y_data()
        self.sd = self.get_sd()
        self.sem = self.get_sem()
        self.provided_noise = self.get_provided_noise()  # in parent class

    def get_mean_y_data(self):
        """
        Return the mean of the y-values that should be plotted
        Returns:
            y_data: The y-value
        """
        variable = self.get_y_variable_name()
        y_data = np.mean(self.line_data[variable])
        y_data = y_data + self.y_offset

        return y_data

    def get_sd(self):
        """
        Return the standard deviation of the y-values that should be plotted.
        Returns:
            sd: The standard deviation
        """
        variable = self.get_y_variable_name()
        y_values = self.line_data[variable]
        sd = np.std(y_values)

        return sd

    def get_sem(self):
        """
        Return the standard error of the mean of the
        y-values that should be plotted.
        Returns:
            sem: The standard error of the mean
        """
        variable = self.get_y_variable_name()
        y_values = self.line_data[variable]
        sem = self.sd / np.sqrt(len(y_values))

        return sem



import numpy as np
import pandas as pd
import petab.C as ptc
import pyqtgraph as pg

from . import bar_row
from . import plot_class


class BarPlot(plot_class.PlotClass):
    """
     Can generate a bar plot based on the given specifications

     Arguments:
         measurement_df: PEtab measurement table
         visualization_df: PEtab visualization table
         simulation_df: PEtab simulation table
         condition_df: PEtab condition table
         plotId: Id of the plot (has to in the visualization_df aswell)

     Attributes:
         bar_rows: A list of BarRows (one for each visualization df row)
         overview_df: A df containing the information of each bar

     """

    def __init__(self, measurement_df: pd.DataFrame = None,
                 visualization_df: pd.DataFrame = None,
                 simulation_df: pd.DataFrame = None,
                 condition_df: pd.DataFrame = None,
                 plotId: str = ""):
        super().__init__(measurement_df, visualization_df, simulation_df,
                         condition_df, plotId)

        self.bar_width = 0.4
        # bar_rows also contains simulation bars
        self.bar_rows = []
        self.add_bar_rows(self.measurement_df)  # list of plot_rows
        self.add_bar_rows(self.simulation_df)

        # A df containing the information needed to plot the bars
        self.overview_df = pd.DataFrame(columns=["x", "y", "name", "sd", "sem", "provided_noise", "is_simulation", "tick_pos"])

        self.plot_everything()

    def plot_everything(self):
        """
        Calculate the overview_df and create the plot based on it.
        If a simulation df is given, also generate the correlation plot.
        """
        self.plot.clear()
        self.overview_df = self.get_bars_df(self.bar_rows)

        self.generate_plot()

        if self.simulation_df is not None:
            # create correlation plot
            self.generate_correlation_plot(self.overview_df)

    def add_bar_rows(self, df):
        """
        Add a BarRow object for each row of the visualization df.

        Arguments:
            df: Measurement or Simulation df
        """
        if self.visualization_df is not None:
            for _, plot_spec in self.visualization_df.iterrows():
                if df is not None:
                    row = bar_row.BarRow(df, plot_spec, self.condition_df)
                    self.bar_rows.append(row)

    def get_bars_df(self, bar_rows):
        """
        Generate a dataframe containing plotting information
        of the individual bars.

        Arguments:
            bar_rows: A list of BarRow objects
        Returns:
            df: A dataframe with information relevant
                for plotting a bar (x, y, sd, etc.)
        """
        bar_rows = [bar_row for bar_row in bar_rows if bar_row.dataset_id not in self.disabled_rows]

        x = range(len(bar_rows))
        tick_pos = range(len(bar_rows))
        y = [bar.y_data for bar in bar_rows]
        names = [bar.legend_name for bar in bar_rows]
        sd = [bar.sd for bar in bar_rows]
        sem = [bar.sem for bar in bar_rows]
        noise = [bar.provided_noise for bar in bar_rows]
        is_simulation = [bar.is_simulation for bar in bar_rows]

        df = pd.DataFrame(list(zip(x, y, names, sd, sem, noise, is_simulation, tick_pos)),
                          columns=["x", "y", "name", "sd", "sem", "provided_noise", "is_simulation", "tick_pos"])

        # Adjust x and tick_pos of the bars when simulation bars are plotted
        # such that they are next to each other
        if self.simulation_df is not None:
            # to keep the order of bars consistent
            indexes = np.unique(df["name"], return_index=True)[1]
            names = [df["name"][index] for index in sorted(indexes)]
            for i, name in enumerate(names):
                # set measurement and simulation bars to same x based on name
                index = df[df["name"] == name].index
                df.loc[index, "x"] = i
                df.loc[index, "tick_pos"] = i

            # separate measurement and simulation bars
            bar_separation_shift = self.bar_width/2
            df.loc[~df["is_simulation"], "x"] -= bar_separation_shift
            df.loc[df["is_simulation"], "x"] += bar_separation_shift

        return df

    def generate_plot(self):
        """
        Generate the plot based on the information in the overview_df df.
        Add error bars to the plot.
        Set y-scale to log10 if necessary.
        """

        if not self.overview_df.empty:
            # get the axis labels info from the first line of the plot
            self.plot.setLabel("left", self.bar_rows[0].y_label)
            self.plot.setLabel("bottom", self.bar_rows[0].x_label)

            # Add bars
            simu_rows = self.overview_df["is_simulation"]
            bar_item = pg.BarGraphItem(x=self.overview_df[~simu_rows]["x"],
                                       height=self.overview_df[~simu_rows]["y"], width=self.bar_width)
            self.plot.addItem(bar_item)  # measurement bars
            bar_item = pg.BarGraphItem(x=self.overview_df[simu_rows]["x"], brush="w",
                                       height=self.overview_df[simu_rows]["y"], width=self.bar_width)
            self.plot.addItem(bar_item)  # simulation bars

            # Add error bars
            error_length = self.overview_df["sd"]
            if self.bar_rows[0].plot_type_data == ptc.MEAN_AND_SEM:
                error_length = self.overview_df["sem"]
            if self.bar_rows[0].plot_type_data == ptc.PROVIDED:
                error_length = self.overview_df["provided_noise"]
            error = pg.ErrorBarItem(x=self.overview_df["x"], y=self.overview_df["y"],
                                    top=error_length, bottom=error_length, beam=0.1)
            self.plot.addItem(error)

            # set tick names to the legend entry of the bars
            xax = self.plot.getAxis("bottom")
            ticks = [list(zip(self.overview_df["tick_pos"], self.overview_df["name"]))]
            xax.setTicks(ticks)

            # set y-scale to log if necessary
            if "log" in self.bar_rows[0].y_scale:
                self.plot.setLogMode(y=True)
                if self.plot_rows[0].x_scale == "log":
                    self.add_warning("log not supported, using log10 instead (in " + self.plot_title + ")")

    def add_or_remove_line(self, dataset_id):
        """
        Add the bar corresponding to the dataset id to the plot
        if it is currently disabled. Otherwise, remove it from the plot.
        Call this method when enabling/disabling rows in
        the visualization df.

        Arguments:
            dataset_id: The id of the bar which should be removed/added
        """
        if dataset_id in self.disabled_rows:
            self.disabled_rows.remove(dataset_id)
        else:
            self.disabled_rows.add(dataset_id)
        # also replots the correlation plot
        self.plot_everything()



import numpy as np
import pandas as pd
import petab.C as ptc
import petab

from . import utils


class RowClass:
    """
    Can add the content of a visualization_df row to a PlotItem.
    Used for line or bar plots.

    Arguments:
        exp_data: PEtab measurement table
        plot_spec: A single row of a PEtab visualization table
        condition_df: PEtab condition table
    Attributes:
        line_data: PEtab measurement or simulation table reduced to relevant rows
        plot_spec: A single row of a PEtab visualization table
        condition_df: PEtab condition table reduced to relevant rows
        dataset_id: Id of the dataset
        x_var: Name of the x-variable
        y_var: Name of the observable that should be plotted
            (Note: It is not the name of the y-variable)
        x_offset: Value of the x-offset
        y_offset: Value of the y-offset
        x_label: Name of the x-axis
        y_label: Name of the y-axis
        x_scale: Scale of the x-axis (linear of log10)
        y_scale: Scale of the y-axis (linear of log10)
        legend_name: Name of the line for the plot-legend
        plot_type_data:  The type how replicates should be handled,
            can be MeanAndSD, MeanAndSEM, replicate or provided
        is_simulation: Boolean, True if exp_data is a simulation df
        has_replicates: Booelean, True if replicates are in line_data
        replicates: List of line_data subsets, divided by replicateId

    """

    def __init__(self, exp_data: pd.DataFrame,
                 plot_spec: pd.Series, condition_df: pd.DataFrame, ):
        self.x_data = []    # placeholder value, will be overwritten by plot_row
        self.y_data = []   # placeholder value, will be overwritten by plot_row/bar_row

        # set attributes
        self.plot_spec = plot_spec
        self.condition_df = condition_df
        self.dataset_id = utils.get_dataset_id(plot_spec)
        self.x_var = utils.get_x_var(plot_spec)
        # Note: y_var is not the name of the y variable
        # but the observable which should be plotted on the y axis
        self.y_var = utils.get_y_var(plot_spec)
        self.x_offset = utils.get_x_offset(plot_spec)
        self.x_label = utils.get_x_label(plot_spec)
        self.y_label = utils.get_y_label(plot_spec)
        self.x_scale = utils.get_x_scale(plot_spec)
        self.y_offset = utils.get_y_offset(plot_spec)
        self.y_scale = utils.get_y_scale(plot_spec)
        self.legend_name = utils.get_legend_name(plot_spec)
        self.plot_type_data = utils.get_plot_type_data(plot_spec)
        self.is_simulation = ptc.SIMULATION in exp_data.columns

        # reduce dfs to relevant rows
        self.line_data = exp_data
        if self.dataset_id and ptc.DATASET_ID in self.line_data:  # != ""
            self.line_data = self.line_data[self.line_data[ptc.DATASET_ID] == self.dataset_id]
        if self.y_var:  # != ""
            # filter by y-values if specified
            self.line_data = self.line_data[self.line_data[ptc.OBSERVABLE_ID] == self.y_var]
        if self.condition_df is not None and self.x_var != ptc.TIME:
            # reduce the condition df to the relevant rows (by condition id)
            self.condition_df = utils.reduce_condition_df(self.line_data, self.condition_df)

        self.has_replicates = petab.measurements.measurements_have_replicates(self.line_data)
        self.replicates = utils.split_replicates(self.line_data)

    def get_data_df(self):
        """
        Represent the data of this row as a dataframe.
        Contains the x- and y-values, the name, the dataset id,
        the name of the x-variable and the boolean is_simulation.
        Note: Each x-/y-value pair gets their own row in the df.

        Returns
            df: The dataframe containing the row information.
        """
        if len(self.x_data) == len(self.y_data):
            df = pd.DataFrame({"x": self.x_data, "y": self.y_data, "name": self.legend_name,
                               "is_simulation": self.is_simulation, "dataset_id": self.dataset_id,
                               "x_label": self.x_label})
            return df
        else:
            raise Exception("Error: The number of x- and y-values are different")

    def get_provided_noise(self):
        """
        Get the provided noise from the noiseParameters column

        Returns:
            The provided noise
        """
        noise = 0
        if self.plot_type_data == ptc.PROVIDED:
            noise = self.line_data[ptc.NOISE_PARAMETERS]
            noise = np.asarray(noise)

        return noise

    def get_y_variable_name(self):
        """
        Get the name of the y-variable

        Returns:
            variable: The name of the y-variable
        """
        return ptc.SIMULATION if self.is_simulation else ptc.MEASUREMENT



import petab.C as ptc
import numpy as np
import pandas as pd
import warnings
from PySide2 import QtCore
from PySide2.QtWidgets import QComboBox


def get_legend_name(plot_spec: pd.Series):
    """
    Returns the plot title of the plot specification
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The name of the legend entry
    """
    legend_name = ""
    if ptc.DATASET_ID in plot_spec.index:
        legend_name = plot_spec[ptc.DATASET_ID]
    if ptc.LEGEND_ENTRY in plot_spec.index:
        legend_name = plot_spec[ptc.LEGEND_ENTRY]

    return legend_name


def get_x_var(plot_spec: pd.Series):
    """
    Returns the name of the x variable of the plot specification
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The name of the x variable
    """
    x_var = "time"  # default value
    if ptc.X_VALUES in plot_spec.index:
        x_var = plot_spec[ptc.X_VALUES]

    return x_var


def get_y_var(plot_spec: pd.Series):
    """
    Returns the observable which should be plotted on the y-axis
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        observable which should be plotted on the y-axis
    """
    y_var = ""  # default value
    if ptc.Y_VALUES in plot_spec.index:
        y_var = plot_spec[ptc.Y_VALUES]

    return y_var


def get_x_offset(plot_spec: pd.Series):
    """
    Returns the x offset
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The x offset
    """
    x_offset = 0  # default value
    if ptc.X_OFFSET in plot_spec.index:
        x_offset = float(plot_spec[ptc.X_OFFSET])

    return x_offset


def get_x_scale(plot_spec: pd.Series):
    """
    Returns the scale of the x axis (lin, log or order)
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The x scale
    """

    x_scale = "lin"  # default value
    if ptc.X_SCALE in plot_spec.index:
        x_scale = plot_spec[ptc.X_SCALE]

    return x_scale


def get_y_scale(plot_spec: pd.Series):
    """
    Returns the scale of the y axis (lin, log or order)
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The x offset
    """
    y_scale = "lin"  # default value
    if ptc.Y_SCALE in plot_spec.index:
        y_scale = plot_spec[ptc.Y_SCALE]

    return y_scale


def get_y_offset(plot_spec: pd.Series):
    """
    Returns the y offset
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The y offset
    """
    y_offset = 0  # default value
    if ptc.Y_OFFSET in plot_spec.index:
        y_offset = float(plot_spec[ptc.Y_OFFSET])

    return y_offset


def get_x_label(plot_spec: pd.Series):
    """
    Returns the label of the x axis
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The label of the x axis
    """
    x_label = get_x_var(plot_spec)  # defaults to x_var
    if ptc.X_LABEL in plot_spec.index:
        x_label = plot_spec[ptc.X_LABEL]

    return x_label


def get_y_label(plot_spec: pd.Series):
    """
    Returns the label of the y axis
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The label of the y axis
    """
    y_label = ptc.MEASUREMENT  # defaults to y_var
    if ptc.Y_LABEL in plot_spec.index:
        y_label = plot_spec[ptc.Y_LABEL]

    return y_label


def get_dataset_id(plot_spec: pd.Series):
    """
    Returns the dataset id
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The dataset id
    """
    dataset_id = ""
    if ptc.DATASET_ID in plot_spec.index:
        dataset_id = plot_spec[ptc.DATASET_ID]

    return dataset_id


def get_plot_type_data(plot_spec: pd.Series):
    """
    Returns the dataset id
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The dataset id
    """
    plot_type_data = "MeanAndSD"
    if ptc.PLOT_TYPE_DATA in plot_spec.index:
        plot_type_data = plot_spec[ptc.PLOT_TYPE_DATA]

    return plot_type_data


def reduce_condition_df(line_data, condition_df):
    """
    Reduce the condition df to the relevant rows based
    on the unique condition ids in the line_data df

    Arguments:
        line_data: A subset of a measurement df
        condition_df: The condition df
    Returns:
        The reduced condition df
    """
    uni_condition_id, uind = np.unique(
        line_data[ptc.SIMULATION_CONDITION_ID],
        return_index=True)
    # keep the ordering which was given by user from top to bottom
    # (avoid ordering by names '1','10','11','2',...)'
    uni_condition_id = uni_condition_id[np.argsort(uind)]

    # extract conditions (plot input) from condition file
    ind_cond = condition_df.index.isin(uni_condition_id)
    condition_df = condition_df[ind_cond]
    return condition_df

def get_plot_title(visualization_df_rows: pd.DataFrame):
    """
    Returns the title of the plot
    Arguments:
       plot_spec: A single row of a visualization df
    Returns:
        The plot title
    """
    plot_title = ""
    if visualization_df_rows is not None:
        if ptc.PLOT_NAME in visualization_df_rows.columns:
            plot_title = visualization_df_rows.iloc[0][ptc.PLOT_NAME]
        elif ptc.PLOT_ID in visualization_df_rows.columns:
            plot_title = visualization_df_rows.iloc[0][ptc.PLOT_ID]

    return plot_title


def mean_replicates(line_data: pd.DataFrame, x_var: str = ptc.TIME,
                    y_var: str = ptc.MEASUREMENT):
    """
    Calculate the mean of the replicates.
    Note: The line_data already has to be reduced to the relevant
        simulationConditionIds for concentration plots

    Arguments:
        line_data: A subset of the measurement file
        x_var: Name of the x-variable
        y_var: Name of the y-variable (measurement or simulation)
    Returns:
        The mean grouped by x_var
    """
    grouping = ptc.TIME
    if x_var != ptc.TIME:
        # for concentration plots we group by
        # simulationConditionId
        grouping = ptc.SIMULATION_CONDITION_ID
    line_data = line_data[[y_var, grouping]]
    means = line_data.groupby(grouping).mean()
    means = means[y_var].to_numpy()

    return means


def sd_replicates(line_data: pd.DataFrame, x_var: str, is_simulation: bool):
    """
    Calculate the standard deviation of the replicates.

    Arguments:
        line_data: A subset of the measurement file
        x_var: Name of the x-variable
        is_simulation: Boolean to check if the y variable
            is measurement or simulation
    Returns:
        The std grouped by x_var
    """
    y_var = ptc.MEASUREMENT
    if is_simulation:
        y_var = ptc.SIMULATION

    grouping = ptc.TIME
    if x_var != ptc.TIME:
        # for concentration plots we group by
        # simulationConditionId
        grouping = ptc.SIMULATION_CONDITION_ID

    line_data = line_data[[grouping, y_var]]
    # std with ddof = 0 (degrees of freedom)
    # to match np.std that is used in petab
    sds = line_data.groupby(grouping).std(ddof=0)
    sds = sds[y_var].to_numpy()
    return sds


def sem_replicates(line_data: pd.DataFrame, x_var: str, is_simulation: bool):
    """
    Calculate the standard error of the mean of the replicates

    Arguments:
        line_data: A subset of the measurement file
        x_var: Name of the x-variable
        is_simulation: Boolean to check if the y variable
            is measurement or simulation
    Returns:
        The std grouped by x_var
    """
    grouping = ptc.TIME
    if x_var != ptc.TIME:
        # for concentration plots we group by
        # simulationConditionId
        grouping = ptc.SIMULATION_CONDITION_ID

    sd = sd_replicates(line_data, x_var, is_simulation)
    n_replicates = [len(replicates) for replicates in line_data.groupby(grouping)]
    sem = sd / np.sqrt(n_replicates)
    return sem


def split_replicates(line_data: pd.DataFrame):
    """
    Split the line_data df into replicate dfs based on their
    replicate Id.
    If no replicateId column is in the line_data, line_data will
    be returned.

    Arguments:
        line_data: A subset of the measurement file
    Returns:
        The std grouped by x_var
    """
    replicates = []
    if ptc.REPLICATE_ID in line_data.columns:
        for repl_id in np.unique(line_data[ptc.REPLICATE_ID]):
            repl = line_data[line_data[ptc.REPLICATE_ID] == repl_id]
            replicates.append(repl)
    else:
        replicates.append(line_data)
    return replicates


def add_plotnames_to_cbox(exp_data: pd.DataFrame, visualization_df: pd.DataFrame, cbox: QComboBox):
    """
    Add the name of every plot in the visualization df
    to the cbox

    Arguments:
        visualization_df: PEtab visualization table
        cbox:  The list of plots (UI)
    """
    if visualization_df is not None:
        plot_ids = np.unique(visualization_df[ptc.PLOT_ID])
        if ptc.PLOT_NAME in visualization_df.columns:

            # to keep the order of plotnames consistent with the plots that are shown
            # for every identical plot_id, the plot_name has to be the same
            indexes = np.unique(visualization_df[ptc.PLOT_ID], return_index=True)[1]
            plot_names = [visualization_df[ptc.PLOT_NAME][index] for index in sorted(indexes)]
            if len(plot_ids) != len(plot_names):
                warnings.warn("The number of plot ids should be the same as the number of plot names")

            for name in plot_names:
                cbox.addItem(name)
        else:
            for id in np.unique(visualization_df[ptc.PLOT_ID]):
                cbox.addItem(id)
    else:
        # the default plots are grouped by observable ID
        # to keep the order of plots consistent with names from the plot selection
        indexes = np.unique(exp_data[ptc.OBSERVABLE_ID], return_index=True)[1]
        observable_ids = [exp_data[ptc.OBSERVABLE_ID][index] for index in sorted(indexes)]
        for observable_id in observable_ids:
            cbox.addItem(observable_id)


def get_signals(source):
    """
    Print out all signals that are implemented in source
    (only for debug purposes)
    """
    cls = source if isinstance(source, type) else type(source)
    signal = type(QtCore.Signal())
    print("Signals:")
    for name in dir(source):
        try:
            if isinstance(getattr(cls, name), signal):
                print(name)
        except Exception as inst:
            print("skipped")



"""Setuptools module for petabvis"""

import pathlib

from setuptools import setup

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    long_description=long_description,
    long_description_content_type='text/markdown',
)
